#include <converter.h>

/*! Constructor of Converter.*/
Converter::Converter(){
    init();
}

/*! Initializes the converter by reading in data files and tables.*/
void Converter::init(){
    QString radiiFile(QCoreApplication::applicationDirPath() + "/msms/atmtypenumbers.txt");
    QFile file(radiiFile);
    if (!file.open(QFile::ReadOnly | QFile::Text)) return;
    QTextStream stream(&file);
    QString line;

    radii = new AtomRadii();
    atoms = new Atoms();

    while(stream.readLineInto(&line)){
        if(line.startsWith("#")) continue;

        line.replace("\t", " ");
        QStringList pieces = line.split(" ");
        pieces.removeAll("");
        int indexHashtag = pieces.indexOf("#");
        if(indexHashtag != -1) pieces = pieces.mid(0, indexHashtag);

        if(pieces.length() == 0) continue;

        if(line.startsWith("radius", Qt::CaseInsensitive)){
            radii->addRadius(pieces.at(1).toInt(), pieces.at(2).toFloat(), pieces.at(3).toFloat(), pieces.length() > 4 ? pieces.at(4).toFloat() : pieces.at(3).toFloat());
            continue;
        }

        QString first = pieces.at(0);

        QString second = pieces.at(1);
        second.replace("_", " ");
        second.replace(".*", "*");
        second.replace("//", "/");

        int atmnum = pieces.at(2).toInt();
        // there is no radius for this element, so create one
        if(!radii->containsAtomID(atmnum)){
            radii->addRadius(atmnum, 0.01f, 0.01f, 0.01f);
        }
        atoms->addAtom(first, second, atmnum, radii->getAtomRadius(atmnum));
    }

    QString hydroFile(QCoreApplication::applicationDirPath() + "/msms/hydrophobicity.txt");
    QFile hFile(hydroFile);
    if (!hFile.open(QFile::ReadOnly | QFile::Text)) return;
    QTextStream hStream(&hFile);

    while(hStream.readLineInto(&line)){
        _hydrophobicity.insert(line.mid(0,3), line.mid(4,8).replace(" ", "").toInt());
    }

}

/*! Constructor a .pdb file to a .xyzrn file while additionally saving information from the .pdf file.*/
int Converter::convert_PDBtoXYZRN(QString &filename, QVector<float>* temperatures, QVector<QString>* bindingSites, QVector<QString>* atomNames, QVector<QString>* resNames, QVector<int>* aminoAcid, QMap<QString, int>* hydrophobicity, QVector<QVector<int>>* connectivity){
    int atomCounter = 0;
    for(auto h : _hydrophobicity.keys())
        hydrophobicity->insert(h, _hydrophobicity.value(h));

    QFile file_in(filename);
    if (!file_in.open(QFile::ReadOnly | QFile::Text)) return 0;
    QTextStream stream_in(&file_in);
    QString line;
    int aminoCount = 0;

    QFileInfo info(filename);
    QFile file_out(QCoreApplication::applicationDirPath() + "/temp/" + info.completeBaseName() + ".xyzrn");
    if (file_out.open(QIODevice::ReadWrite)) {
        QTextStream stream_out(&file_out);

        bool moreAtoms = true;
        while(stream_in.readLineInto(&line)){
            if(line.startsWith("ATOM", Qt::CaseInsensitive) || line.startsWith("HETATM", Qt::CaseInsensitive)){
                if(!moreAtoms) continue;
                QString aname = line.mid(12,4).replace(" ", "");
                QString resname = line.mid(17,3).replace(" ", "");
                QString resnum = line.mid(22,4).replace(" ", "");
                QString x = line.mid(30,8);
                QString y = line.mid(38,8);
                QString z = line.mid(46,8);
                QString occupancy = line.mid(54,6);
                QString temperature = line.mid(60,6);
                temperatures->push_back(temperature.toFloat());
                if(!atomNames->contains(aname))
                    atomNames->push_back(aname);
                if(!resNames->contains(resname))
                    resNames->push_back(resname);
                aminoAcid->push_back(aminoCount);

                if(aname.at(0).isNumber() && (aname.at(1) == "H" || aname.at(1) == "h" || aname.at(1) == "D" || aname.at(1) == "d")){
                    aname = QString("H");
                }else if((aname.at(0) == "H" || aname.at(0) == "h") && (aname.at(1) != "G" && aname.at(1) != "g")){
                    aname = QString("H");
                }

                Converter::Atoms::Atom atom = atoms->getAtom(aname, resname);
                stream_out << x << " " << y << " " << z << " " << atom.radius.radiusUnited << " " << "1" << " " << aname << "_" << resname << "_" << resnum <<"\n";
                atomCounter++;
            }else if(line.startsWith("LINK", Qt::CaseInsensitive)){
                QString aname = line.mid(12,4).replace(" ", "");
                QString resname = line.mid(17,3).replace(" ", "");
                QString resnum = line.mid(22,4).replace(" ", "");
                bindingSites->push_back(aname+"_"+resname+"_"+resnum);
                aname = line.mid(42,4).replace(" ", "");
                resname = line.mid(47,3).replace(" ", "");
                resnum = line.mid(52,4).replace(" ", "");
                bindingSites->push_back(aname+"_"+resname+"_"+resnum);
            }else if(line.startsWith("TER", Qt::CaseInsensitive)){
                if(!moreAtoms) continue;
                aminoCount++;
            }else if(line.startsWith("ENDMDL", Qt::CaseInsensitive)){
                moreAtoms = false;
            }else if(line.startsWith("CONECT", Qt::CaseInsensitive)){
                QVector<int> temp;
                QString l;
                l = line.mid(6,5).replace(" ", "");
                if(l.length() != 0)
                    temp.push_back(l.toInt());

                l = line.mid(11,5).replace(" ", "");
                if(l.length() != 0)
                    temp.push_back(l.toInt());

                l = line.mid(16,5).replace(" ", "");
                if(l.length() != 0)
                    temp.push_back(l.toInt());

                l = line.mid(21,5).replace(" ", "");
                if(l.length() != 0)
                    temp.push_back(l.toInt());

                l = line.mid(26,5).replace(" ", "");
                if(l.length() != 0)
                    temp.push_back(l.toInt());

                connectivity->push_back(temp);
            }
        }

        file_out.close();
    }
    filename = file_out.fileName();
    return atomCounter;
}

/*! Add a new radius to the data structure.*/
void Converter::AtomRadii::addRadius(int n, float dist, float radiusExplicit, float radiusUnited){
    radii.push_back(AtomRadius(n, dist, radiusExplicit, radiusUnited));
}

/*! Check if the current atom is contained in the radii.*/
bool Converter::AtomRadii::containsAtomID(int n){

    for(auto iter = radii.begin(); iter != radii.end(); ++iter){
        if((*iter).n == n) return true;
    }
    return false;
}

/*! Returns the atom radius for the ID.*/
Converter::AtomRadii::AtomRadius Converter::AtomRadii::getAtomRadius(int n){
    for(auto iter = radii.begin(); iter != radii.end(); ++iter){
        if((*iter).n == n) return (*iter);
    }
}

/*! Adds an atom to the structure.*/
void Converter::Atoms::addAtom(QString respat, QString atmpat, int atmnum, AtomRadii::AtomRadius radius){
    atoms.push_back(Atom(respat, atmpat, atmnum, radius));
}

/*! Compares two Atom entries.*/
bool Converter::Atoms::compareAtomEntries(QString nameA, QString nameB){
    if(nameA == nameB) return true;
    if(nameB == "*") return true;
    if(nameB.contains("|")){
        QStringList list = nameB.split("|");
        for(int i = 0; i < list.length(); i++){
            if(compareAtomEntries(nameA, list.at(i))) return true;
        }
        return false;
    }
    if(nameB.length() < nameA.length()) return false;

    int offset = 0;
    for(int i = 0; i < nameA.length(); i++){
        QChar current = nameA.at(i);
        int i_mod = i + offset;
        if(nameB.length() <= i_mod) return false;
        if(current != nameB.at(i_mod) && nameB.at(i_mod) != '*' && nameB.at(i_mod) != '.'){
            if(nameB.at(i_mod) == "["){
                int start = i_mod;
                int end = nameB.indexOf("]", start);
                bool inRegion = false;
                if(nameB.indexOf("-", start) < end){
                    for(int j = nameB.at(start+1).unicode(); j <= nameB.at(end-1).unicode(); j++){
                        if(current.unicode() == j) inRegion = true;
                    }
                }else{
                    for(int j = start + 1; j <= end-1; j++){
                        if(current == nameB.at(j)) inRegion = true;
                    }
                }
                if(inRegion){
                    offset += end-start;
                    continue;
                }else{
                    return false;
                }
            }
            return false;
        }
    }

    if(nameA.length() + offset != nameB.length()){
        if(nameB.at(nameA.length() + offset) == "["){
            int start = nameA.length() + offset;
            int end = nameB.indexOf("]", start);
            if(nameB.length() < end+2) return false;
            if(nameB.at(end+1) == "?") return true;
        }
        return false;
    }
    return true;
}


/*! Returns an atom that fits to the given data.*/
Converter::Atoms::Atom Converter::Atoms::getAtom(QString aname, QString resname){
    for(auto iter = atoms.begin(); iter != atoms.end(); ++iter){
        if(compareAtomEntries(aname, (*iter).atmpat) && compareAtomEntries(resname, (*iter).respat)){
            return (*iter);
        }
    }
    return Atom("", "", -1, Converter::AtomRadii::AtomRadius(-1, 0.01f, 0.01f, 0.01f));
}

/*! Returns the number of atoms.*/
int Converter::Atoms::getAtomCount(){
    return atoms.size();
}
