#include "mainwindow.h"
#include "ui_mainwindow.h"

/*! Constructor of MainWindow.*/
MainWindow::MainWindow(QWidget *parent) :
  QMainWindow(parent),
  ui(new Ui::MainWindow)
{
  ui->setupUi(this);
  ui->actionCombined_Rotation->setChecked(true);

  converter = new Converter();
  colorSelector = new ColorSelector(this);

  connect(this, SIGNAL(loadMesh(QString, QVector<float>*, QVector<QString>*, QVector<QString>*, QVector<QString>*, QVector<int>*, QMap<QString, int>*, QVector<QVector<int>>*)), ui->MeshView, SLOT(loadMesh(QString, QVector<float>*, QVector<QString>*, QVector<QString>*, QVector<QString>*, QVector<int>*, QMap<QString, int>*, QVector<QVector<int>>*)));
  connect(ui->MeshView, SIGNAL(newMeshLoaded(SurfaceMesh*)), ui->MeshSphereView, SLOT(newMeshLoaded(SurfaceMesh*)));

  connect(this, SIGNAL(combinedRotationToggle(bool)), ui->MeshView, SLOT(combinedRotationToggle(bool)));
  connect(this, SIGNAL(combinedRotationToggle(bool)), ui->MeshSphereView, SLOT(combinedRotationToggle(bool)));

  connect(ui->MeshView, SIGNAL(modelMatrixChangedMesh(QMatrix4x4)), ui->MeshSphereView, SLOT(modelMatrixChangedMesh(QMatrix4x4)));
  connect(ui->MeshSphereView, SIGNAL(modelMatrixChangedMeshSphere(QMatrix4x4)), ui->MeshView, SLOT(modelMatrixChangedMeshSphere(QMatrix4x4)));

  connect(ui->MeshView, SIGNAL(renderModeChangedMesh(int)), ui->MeshSphereView, SLOT(renderModeChangedMesh(int)));
  connect(ui->MeshSphereView, SIGNAL(renderModeChangedMeshSphere(int)), ui->MeshView, SLOT(renderModeChangedMeshSphere(int)));  

  connect(colorSelector, SIGNAL(changeColor(int)), ui->MeshView, SLOT(changeColor(int)));
  connect(ui->MeshView, SIGNAL(colorChanged(QVector<QVector3D>*)), ui->MeshSphereView, SLOT(colorChanged(QVector<QVector3D>*)));

  connect(ui->MeshSphereView, SIGNAL(surfaceMeshSphereChanged(SurfaceMeshSphere *)), ui->TextureView, SLOT(surfaceMeshSphereChanged(SurfaceMeshSphere *)));

  QWidget *spacerWidget = new QWidget(this);
  spacerWidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
  spacerWidget->setVisible(true);
  ui->mainToolBar->addWidget(spacerWidget);
  progressBar = new QProgressBar();
  progressBar->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Preferred);
  ui->mainToolBar->addWidget(progressBar);

}

/*! Destructor of MainWindow.*/
MainWindow::~MainWindow()
{
  delete ui;
}

/*! Handles the close event.*/
void MainWindow::closeEvent(QCloseEvent *event) {
    QString geometryFilepath = QCoreApplication::applicationDirPath() + "/temp";
    QDir path(geometryFilepath);
    if(path.exists()){
        path.removeRecursively();
    }
    QApplication::closeAllWindows();
    event->accept();
}

/*! Handles the action of a chosen file.*/
void MainWindow::on_actionLoad_File_triggered()
{
  QString filename = QFileDialog::getOpenFileName(this, tr("Open File"), QCoreApplication::applicationDirPath(), tr("Data Files (*.pdb)"));
  if(filename.isEmpty()) {
    return;
  }else{
    QProgressDialog progress("Converting file...", "", 0, 3, this);
    progress.setCancelButton(0);
    progress.setWindowFlags(Qt::SplashScreen);
    progress.setWindowModality(Qt::WindowModal);
    progress.show();
    QApplication::processEvents();

    QString geometryFilepath = QCoreApplication::applicationDirPath() + "/temp";
    if(!QDir(geometryFilepath).exists()){
      QDir().mkdir(geometryFilepath);
    }

    QVector<float>* temperatures = new QVector<float>;
    QVector<QString>* bindingSites = new QVector<QString>;
    QVector<int>* aminoAcid = new QVector<int>;
    QVector<QString>* atomNames = new QVector<QString>;
    QVector<QString>* resNames = new QVector<QString>;
    QMap<QString, int>* hydrophobicity = new QMap<QString, int>;
    QVector<QVector<int>>* connectivity = new QVector<QVector<int>>;
    int atomCounter = converter->convert_PDBtoXYZRN(filename, temperatures, bindingSites, atomNames, resNames, aminoAcid, hydrophobicity, connectivity);
    progress.setLabelText("Generating surface mesh...");
    progress.setValue(1);

    QFileInfo info(filename);
    QString geometryFilename = geometryFilepath + "/" + info.baseName();

    if(!(QFile(geometryFilename + ".vert").exists() && QFile(geometryFilename + ".face").exists())){
        qDebug() << "Number of Atoms: " << atomCounter;

        QStringList arguments = QStringList() << "-if" << filename << "-of" << geometryFilename << "-density" << QString::number(3.0) << "-no_header" ;
        QProcess *process = new QProcess(this);
        process->start("msms/msms.exe", arguments);

        if (process->waitForFinished()){
            if (process->readAll().contains("MSMS terminated normally")){
                progress.setLabelText("Loading mesh...");
                progress.setValue(2);
                meshLoaded = true;
                emit loadMesh(geometryFilename, temperatures, bindingSites, atomNames, resNames, aminoAcid, hydrophobicity, connectivity);
            }else{
                progress.setValue(3);
                QMessageBox error;
                error.setWindowFlags(Qt::SplashScreen);
                error.setText("An error occured while generating the surface mesh.");
                error.exec();
            }
        }
    }else{
        progress.setLabelText("Loading mesh...");
        progress.setValue(2);
        meshLoaded = true;
        emit loadMesh(geometryFilename, temperatures, bindingSites, atomNames, resNames, aminoAcid, hydrophobicity, connectivity);
    }

  }
}

/*! Handles the action of a new selected color.*/
void MainWindow::on_actionColor_Scheme_triggered()
{
    if(meshLoaded){
        colorSelector->show();
    }else{
        QMessageBox error;
        error.setWindowFlags(Qt::SplashScreen);
        error.setText("Load a file first.");
        error.exec();
    }
}

/*! Handles the action of toggling the combined rotation.*/
void MainWindow::on_actionCombined_Rotation_triggered(){
    combinedRotation = !combinedRotation;
    emit combinedRotationToggle(combinedRotation);
}

/*! Handles the update of the progress bar.*/
void MainWindow::updateProgress(int progress){
    progressBar->setValue(progress);
}

/*! Handles the toogle of the UI.*/
void MainWindow::toggleUI(bool toggle){
    if(toggle){
        ui->actionLoad_File->setEnabled(false);
        ui->actionColor_Scheme->setEnabled(false);
        ui->actionCombined_Rotation->setEnabled(false);
    }else{
        ui->actionLoad_File->setEnabled(true);
        ui->actionColor_Scheme->setEnabled(true);
        ui->actionCombined_Rotation->setEnabled(true);
    }
}


