#include "meshsphereview.h"

/*! Constructor of meshSphereView. */
meshSphereView::meshSphereView(QWidget *parent) : QOpenGLWidget(parent){
  QSurfaceFormat f = QSurfaceFormat::defaultFormat();
  QSysInfo sysinfo;
  if(sysinfo.productType().toCaseFolded() == QString("osx") ||
     sysinfo.productType().toCaseFolded() == QString("macos")){
      f.setVersion( 3, 3 );
      f.setProfile(QSurfaceFormat::CoreProfile);
  }
  setFormat(f);
  QSurfaceFormat::setDefaultFormat(f);

  _modelMatrix.setToIdentity();
  _viewMatrix.setToIdentity();
  _projectionMatrix.setToIdentity();
  _projectionMatrix.perspective(45.0f, this->width() / this->height(), 0.1f, 500.0f);

  _currentMesh = new SurfaceMeshSphere();
  connect(_currentMesh, SIGNAL(updateProgress(int)), parent->parent()->parent()->parent(), SLOT(updateProgress(int)));
  connect(this, SIGNAL(toggleUI(bool)), parent->parent()->parent()->parent(), SLOT(toggleUI(bool)));

  this->setFocusPolicy(Qt::StrongFocus);
  this->setMouseTracking(true);
}

/*! Initialize GL. */
void meshSphereView::initializeGL(){
  initializeOpenGLFunctions();
  _gl = this->context()->extraFunctions();

  if(!_m_program.addShaderFromSourceFile(QOpenGLShader::Vertex, ":/shaders/meshSphere.vert") ||
     !_m_program.addShaderFromSourceFile(QOpenGLShader::Fragment, ":/shaders/meshSphere.frag") ||
     !_m_program.link()) {
    qDebug() << _m_program.log();
  }
  _currentMesh->initGL(_gl);
  glClearColor(0.8, 0.8, 0.8, 1);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
}

/*! Paint the current content. */
void meshSphereView::paintGL(){
    if(_currentMesh->isLoaded()){
        if(_renderMode == 1){
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        }else if(_renderMode == 2){
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        }else if(_renderMode == 3){
            glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
        }

        _m_program.bind();
        _m_program.setUniformValue("renderMode", _renderMode);
        _m_program.setUniformValue("modelMatrix", _modelMatrix);
        _m_program.setUniformValue("viewMatrix", _viewMatrix);
        _m_program.setUniformValue("projectionMatrix", _projectionMatrix);
        _currentMesh->draw();
        _m_program.release();
    }
}

/*! Resize the window. */
void meshSphereView::resizeGL(int w, int h){
    _projectionMatrix.setToIdentity();
    _projectionMatrix.perspective(45.0f, qreal(w) / qreal(h), 0.1, 500.0);
}

/*! Handle a wheel event. */
void meshSphereView::wheelEvent(QWheelEvent *event){
    QMatrix4x4 translationMatrix;
    translationMatrix.setToIdentity();
    translationMatrix.translate(0.0f, 0.0f, event->delta() / 25.0f);
    _viewMatrix = translationMatrix * _viewMatrix;
    update();
}


/*! Handle a mouse press event. */
void meshSphereView::mousePressEvent(QMouseEvent *event){
    _lastPos = event->pos();
}

/*! Handle a mouse move event. */
void meshSphereView::mouseMoveEvent(QMouseEvent *event){


    if (event->buttons() & Qt::LeftButton) {
        float dx = event->x() - _lastPos.x();
        float dy = event->y() - _lastPos.y();
        _lastPos = event->pos();

        QMatrix4x4 transformationMatrix;
        transformationMatrix.setToIdentity();
        QVector3D vector(-dy, dx, 0.0);
        vector.normalize();
        transformationMatrix.translate(_currentMesh->getCenter());
        transformationMatrix.rotate(5.0f, vector);
        transformationMatrix.translate(-_currentMesh->getCenter());
        _modelMatrix = transformationMatrix * _modelMatrix;

        emit modelMatrixChangedMeshSphere(_modelMatrix);
    }else if(event->buttons() & Qt::RightButton){
        float dx = event->x() - _lastPos.x();
        float dy = event->y() - _lastPos.y();
        _lastPos = event->pos();
        QMatrix4x4 transformationMatrix;
        transformationMatrix.setToIdentity();
        QVector3D vector(dx/12.5, -dy/12.5, 0.0);
        transformationMatrix.translate(vector);
        _viewMatrix = transformationMatrix * _viewMatrix;
    }
    update();
}

/*! Handle a key press event. */
void meshSphereView::keyPressEvent(QKeyEvent *ke){
    GLfloat amount = 5.0f;
    QVector3D translation;
    switch (ke->key())
    {
        case Qt::Key_W:
            translation = QVector3D(0.0f, 0.0f, amount);
            break;
        case Qt::Key_S:
            translation = QVector3D(0.0f, 0.0f, -amount);
            break;
        case Qt::Key_D:
            translation = QVector3D(-amount, 0.0f, 0.0f);
            break;
        case Qt::Key_A:
            translation = QVector3D(amount, 0.0f, 0.0f);
            break;
        case Qt::Key_Q:
            translation = QVector3D(0.0f, amount, 0.0f);
            break;
        case Qt::Key_E:
            translation = QVector3D(0.0f, -amount, 0.0f);
            break;
        case Qt::Key_1:
            _renderMode = 1;
            emit renderModeChangedMeshSphere(_renderMode);
            break;
        case Qt::Key_2:
            _renderMode = 2;
            emit renderModeChangedMeshSphere(_renderMode);
            break;
        case Qt::Key_3:
            _renderMode = 3;
            emit renderModeChangedMeshSphere(_renderMode);
            break;
    }
    QMatrix4x4 translationMatrix;
    translationMatrix.setToIdentity();
    translationMatrix.translate(translation);
    _viewMatrix = translationMatrix * _viewMatrix;
    update();
}

/*! Changes the color to the selected index. */
void meshSphereView::colorChanged(QVector<QVector3D>* colors){
    _currentMesh->changeColor(*colors);
}

/*! Start creating the sphere in another thread. */
void meshSphereView::newMeshLoaded(SurfaceMesh *mesh){
    emit toggleUI(true);

    _renderMode = 1;
    _modelMatrix.setToIdentity();

    QVector3D c = mesh->getCenter();
    float r = mesh->getRadius();

    _cameraPosition = QVector3D(c.x(), c.y(), c.z() - qMin(3.5f * r, 250.0f));
    _cameraTarget = c;
    _viewMatrix.setToIdentity();
    _viewMatrix.lookAt(_cameraPosition, _cameraTarget, QVector3D(0.0f, 1.0f, 0.0f));

   _currentMesh->set(mesh->getVertices(), mesh->getNormals(), mesh->getColors(), mesh->getIndices(), mesh->getCenter(), mesh->getRadius(), mesh->getBarycenter());

   QFuture<void> future = QtConcurrent::run(_currentMesh, &SurfaceMeshSphere::convertToSphere);
   QFutureWatcher<void> *watcher = new QFutureWatcher<void>(this);
   connect(watcher, SIGNAL(finished()), this, SLOT(sphereGenerated()));
   connect(watcher, SIGNAL(finished()), watcher, SLOT(deleteLater()));
   watcher->setFuture(future);
}

/*! After the loading is finished, the data is further processed. */
void meshSphereView::sphereGenerated(){
    _m_program.bind();
    _currentMesh->updateVerticesAndNormals();
    _m_program.release();

    emit surfaceMeshSphereChanged(this->_currentMesh);
    update();
    emit toggleUI(false);
}

/*! If enabled, takes the input modelMatrix as its own. */
void meshSphereView::modelMatrixChangedMesh(QMatrix4x4 modelMatrix)
{
    if(!_combinedRotation) return;
    _modelMatrix = modelMatrix;
    update();
}

/*! Takes the input renderMode as its own. */
void meshSphereView::renderModeChangedMesh(int renderMode)
{
    _renderMode = renderMode;
    update();
}

/*! Toggles the combined rotation. */
void meshSphereView::combinedRotationToggle(bool toggle){
    _combinedRotation = toggle;
}
