#include "surfacemeshsphere.h"

/*! Constructor of SurfaceMeshSphere.*/
SurfaceMeshSphere::SurfaceMeshSphere(){ }

/*! Destructor of SurfaceMeshSphere. */
SurfaceMeshSphere::~SurfaceMeshSphere(){
    _vertexBuffer.destroy();
    _normalBuffer.destroy();
    _colorBuffer.destroy();
    _indexBuffer.destroy();
    _gl->glDeleteVertexArrays(1, &_vao);
}

/*! Initializes the GL. */
void SurfaceMeshSphere::initGL(QOpenGLExtraFunctions *gl){
    _gl = gl;
}

/*! Sets the current mesh with all data given. This method is used when a new molecule is loaded. */
void SurfaceMeshSphere::set(QVector<QVector3D>* vertices, QVector<QVector3D>* normals, QVector<QVector3D>* colors, QVector<GLuint>* indices, QVector3D center, float radius, QVector3D barycenter){
    _vertices = *vertices;
    _normals = *normals;
    _colors = *colors;
    _indices = *indices;
    _center = center;
    _radius = radius;
    _barycenter = barycenter;

    //Bind VAO
    _gl->glGenVertexArrays(1, &_vao);
    _gl->glBindVertexArray(_vao);

    //Bind VBO
    //Bind vertexBuffer
    _vertexBuffer.create();
    _vertexBuffer.bind();
    _vertexBuffer.allocate(_vertices.data(), static_cast<int>(_vertices.size()) * sizeof(QVector3D));

    //Bind normalBuffer
    _normalBuffer.create();
    _normalBuffer.bind();
    _normalBuffer.allocate(_normals.data(), static_cast<int>(_normals.size()) * sizeof(QVector3D));

    //Bind colorBuffer
    _colorBuffer.create();
    _colorBuffer.bind();
    _colorBuffer.allocate(_colors.data(), static_cast<int>(_colors.size()) * sizeof(QVector3D));

    //Bind indexBuffer
    QOpenGLBuffer inedxBufferProxy(QOpenGLBuffer::IndexBuffer);
    inedxBufferProxy.create();
    _indexBuffer = inedxBufferProxy;
    _indexBuffer.bind();
    _indexBuffer.allocate(_indices.data(), static_cast<int>(_indices.size()) * sizeof(GLuint));

    loaded = true;

    _gl->glBindVertexArray(0);
    _vertexBuffer.release();
    _normalBuffer.release();
    _colorBuffer.release();
    _indexBuffer.release();
}

/*! Spherical to cartesian helper function. */
QVector3D ToCartesian(float theta, float phi)
{
    QVector3D cartesian;
    cartesian.setX(sin(theta)*cos(phi));
    cartesian.setY(sin(theta)*sin(phi));
    cartesian.setZ(cos(theta));
    return cartesian;
}

/*! Converts the stored mesh data to a sphere. */
void SurfaceMeshSphere::convertToSphere(){
    emit updateProgress(0);
    loaded = false;
    float minErr = 1.0 / _vertices.size();

    /*
     * Find all unique edges from the indices
     */
    std::map<QString, Edge> edgesVertices;
    for(int i = 0; i < _indices.size(); i+=3){
        QString id;
        Edge e;

        if(_indices[i] > _indices[i+1]){
            id = QString(_indices[i+1])+QString(_indices[i]);
            e.v1 = _indices[i+1];
            e.v2 = _indices[i];
        }else{
            id = QString(_indices[i])+QString(_indices[i+1]);
            e.v1 = _indices[i];
            e.v2 = _indices[i+1];
        }

        if (edgesVertices.find(id) == edgesVertices.end())
            edgesVertices[id] = e;


        if(_indices[i+1] > _indices[i+2]){
            id = QString(_indices[i+2])+QString(_indices[i+1]);
            e.v1 = _indices[i+2];
            e.v2 = _indices[i+1];
        }else{
            id = QString(_indices[i+1])+QString(_indices[i+2]);
            e.v1 = _indices[i+1];
            e.v2 = _indices[i+2];
        }

        if (edgesVertices.find(id) == edgesVertices.end())
            edgesVertices[id] = e;


        if(_indices[i+2] > _indices[i]){
            id = QString(_indices[i])+QString(_indices[i+2]);
            e.v1 = _indices[i];
            e.v2 = _indices[i+2];
        }else{
            id = QString(_indices[i+2])+QString(_indices[i]);
            e.v1 = _indices[i+2];
            e.v2 = _indices[i];
        }

        if (edgesVertices.find(id) == edgesVertices.end())
            edgesVertices[id] = e;
    }


    /*
     * Save all edges for every vertex
     */
    QVector<QVector<int>> verticesEdges;
    for(int i = 0; i < _vertices.size(); i++){
        QVector<int> t;
        verticesEdges.push_back(t);
    }
    for (auto const& x : edgesVertices)
    {
        verticesEdges[x.second.v1].push_back(x.second.v2);
        verticesEdges[x.second.v2].push_back(x.second.v1);
    }

    /*
     * Find two vertices that are far away
     */
    int iMin = -1, iMax = -1;
    QVector<float> verticesZ;
    QVector<float> verticesPhi;
    QVector<int> verticesOnBS;

    for(int i = 0; i < _vertices.size(); i++){
        verticesZ.push_back(0);
        verticesPhi.push_back(M_PI);

        float d = _vertices[i].distanceToPoint(_center);
        if(_radius-_radius/10.0 < d && d < _radius+_radius/10.0)
            verticesOnBS.push_back(i);
    }

    float dist = -1;
    for(int i = 0; i < verticesOnBS.size(); i++){
        QVector3D v1 = _vertices[verticesOnBS[i]];
        for(int j = 0; j < verticesOnBS.size(); j++){
            QVector3D v2 = _vertices[verticesOnBS[j]];
            float d = v1.distanceToPoint(v2);
            if(d > dist){
                 dist = d;
                 iMin = verticesOnBS[i];
                 iMax = verticesOnBS[j];
            }
        }
    }


    /*
     * Remove circumpolar vertices and assign Z
     */
    QVector<int> vMinEdges = verticesEdges[iMin];
    verticesEdges[iMin] = {};
    QVector<int> vMaxEdges = verticesEdges[iMax];
    verticesEdges[iMax] = {};

    float minZ = -1.0;
    float maxZ = 1.0;

    verticesZ[iMin] = minZ;
    verticesZ[iMax] = maxZ;

    for(int i = 0; i < vMinEdges.size(); i++){
        verticesZ[vMinEdges[i]] = minZ;
    }
    for(int i = 0; i < vMaxEdges.size(); i++){
        verticesZ[vMaxEdges[i]] = maxZ;
    }

    /*
     * Relax the Z-values
     */
    int M = 5000;
    QVector<QVector<int>> verticesEdgesZ = verticesEdges;
    for(int i = 0; i < M; i++){
        bool breakIter = true;
        for(int j = 0; j < verticesEdgesZ.size(); j++){
            QVector<int> edges = verticesEdgesZ[j];
            if(edges.size() == 0.0) continue;
            if(verticesZ[j] == minZ || verticesZ[j] == maxZ) continue;
            float comb = 0.0f;
            for(int k = 0; k < edges.size(); k++){
                comb += verticesZ[edges[k]];
            }
            float newZ = comb / edges.size();
            float oldZ = verticesZ[j] / M_PI;
            verticesZ[j] = newZ;
            newZ /= M_PI;
            if(abs((newZ - oldZ)/newZ) > qMin(0.025, minErr*1000.0))
                breakIter = false;
        }
        if(breakIter)
            break;

        emit updateProgress((float)i/(float)M * 50);
    }
    emit updateProgress(50);

    /*
     * Find the boundary along the steepest decline
     */
    QVector<int> boundary;
    QVector<int> boundaryEdges;
    boundary.push_back(vMaxEdges[0]);
    QVector<int> neighbours = verticesEdges[vMaxEdges[0]];
    for(int i = 0; i < neighbours.size(); i++){
        if(neighbours[i] == iMin || neighbours[i] == iMax) continue;
        boundaryEdges.push_back(neighbours[i]);
    }
    verticesPhi[vMaxEdges[0]] = 2.0*M_PI;
    float currentZ = maxZ;
    int count = 0;
    while(currentZ != minZ){
        QVector<int> vBoundaryEdges = verticesEdges[boundary[count]];

        float min = maxZ;
        int current = -1;
        for(int i = 0; i < vBoundaryEdges.size(); i++){
            if(vBoundaryEdges[i] == iMax) continue;
            if(verticesZ[vBoundaryEdges[i]] < min){
                min = verticesZ[vBoundaryEdges[i]];
                current = vBoundaryEdges[i];
            }
        }
        count++;
        boundary.push_back(current);
        verticesPhi[current] = 2.0*M_PI;
        neighbours = verticesEdges[current];
        for(int i = 0; i < neighbours.size(); i++){
            if(neighbours[i] == iMin || neighbours[i] == iMax) continue;
            boundaryEdges.push_back(neighbours[i]);
        }
        currentZ = min;
    }

    /*
     * Find the two sets next to the boundary
     */
    QVector<int> setA, setB;
    bool change = true;
    while(change){
        int s = setA.size();
        for(int i = 0; i < boundaryEdges.size(); i++){
            if(boundary.contains(boundaryEdges[i])) continue;

            if(setA.size() == 0)
                setA.push_back(boundaryEdges[i]);

            for(int j = 0; j < setA.size(); j++){
                QVector<int> e = verticesEdges[setA[j]];
                if(e.contains(boundaryEdges[i]) && !setA.contains(boundaryEdges[i])){
                    setA.push_back(boundaryEdges[i]);
                }
            }
        }
        if(s == setA.size())
            change = false;
    }

    for(int i = 0; i < boundaryEdges.size(); i++){
        if(boundary.contains(boundaryEdges[i])) continue;
        if(!setA.contains(boundaryEdges[i]))
            setB.push_back(boundaryEdges[i]);
    }

    /*
     * Relax the Phi-values
     */
    _verticesSphere = _vertices;
    int N = 5000;
    for(int i = 0; i < N; i++){
        bool breakIter = true;
        for(int j = 0; j < verticesEdges.size(); j++){
            QVector<int> edges = verticesEdges[j];
            if(boundary.contains(j)) continue;
            if(edges.size() == 0.0) continue;

            float comb = 0.0f;
            int count = 0;
            for(int k = 0; k < edges.size(); k++){
                if(edges[k] == iMin || edges[k] == iMax) continue;
                if(boundary.contains(edges[k])){
                    if(setA.contains(j)){
                        comb += 2*M_PI;
                    }else if(setB.contains(j)){
                        comb += 0.0;
                    }
                }else{
                    comb += verticesPhi[edges[k]];
                }
                count++;
            }
            float newPhi = comb / count;
            float oldPhi = verticesPhi[j] / (2*M_PI);
            verticesPhi[j] = newPhi;
            newPhi /= (2*M_PI);
            if(abs((newPhi - oldPhi)/newPhi) > minErr)
                breakIter = false;
        }
        if(breakIter)
            break;

        emit updateProgress(50 + (float)i/(float)N * 50);
    }

    /*
     * Create the vertices on the sphere with mercator mapping for the Z-values.
     */
    _uvs.reserve(_vertices.size());
    _longiLati.reserve(_vertices.size());

    float theta_0 = M_PI/(float)(count+1);
    float R = 1.0/(log((cos(theta_0)+1)/(sin(theta_0))));
    for(int i = 0; i < _vertices.size(); i++){
        float theta = 2.0*atan(exp(verticesZ[i]/R));
        if(iMax == i){
            theta = M_PI;
            verticesPhi[i] = 2*M_PI;
        }
        if(iMin == i){
            theta = 0;
            verticesPhi[i] = 0;
        }

        _normals[i] = ToCartesian(theta, verticesPhi[i]);
        _vertices[i] = _center + _normals[i] * _radius;
        _verticesSphere[i] = _center + _normals[i];

        float u = 0.5f + (atan2(_normals[i].z(), _normals[i].x()) / (2 * M_PI));
        float v = 0.5f - (asin(_normals[i].y())/M_PI);
        _uvs.push_back(QVector2D(u, v));
        _longiLati.push_back(QVector2D(theta, verticesPhi[i]));
    }

    /*
     * Flip the Normals if needed
     */
    QVector3D v0 = _vertices[_indices[0]];
    QVector3D v1 = _vertices[_indices[1]];
    QVector3D v2 = _vertices[_indices[2]];
    QVector3D cross = QVector3D::crossProduct(v1-v0, v2-v0).normalized();

    if(QVector3D::dotProduct(cross, _normals[_indices[2]]) < 0){
        for(int i = 0; i < _vertices.size(); i++){
            _normals[i] *= -1.0;
            _vertices[i] = _center + _normals[i] * _radius;
            _verticesSphere[i] = _center + _normals[i];
        }
    }
    emit updateProgress(100);
}

/*! Updates all vertices and normals. */
void SurfaceMeshSphere::updateVerticesAndNormals(){
    _normalsSphere = _normals;

    _vertexBuffer.create();
    _vertexBuffer.bind();
    _vertexBuffer.allocate(_vertices.data(), static_cast<int>(_vertices.size()) * sizeof(QVector3D));
    _vertexBuffer.release();

    _normalBuffer.create();
    _normalBuffer.bind();
    _normalBuffer.allocate(_normals.data(), static_cast<int>(_normals.size()) * sizeof(QVector3D));
    _normalBuffer.release();
    draw();

    loaded = true;
}

/*! Draws the mesh. */
void SurfaceMeshSphere::draw(){
    _gl->glBindVertexArray(_vao);
    _gl->glEnableVertexAttribArray(0);
    _vertexBuffer.bind();
    _gl->glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);

    _gl->glEnableVertexAttribArray(1);
    _normalBuffer.bind();
    _gl->glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);

    _gl->glEnableVertexAttribArray(2);
    _colorBuffer.bind();
    _gl->glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 0, 0);

    _indexBuffer.bind();
    _gl->glDrawElements(GL_TRIANGLES, _indices.size() * 3, GL_UNSIGNED_INT, 0);

    _gl->glDisableVertexAttribArray(0);
    _gl->glDisableVertexAttribArray(1);
    _gl->glDisableVertexAttribArray(2);
    _gl->glBindVertexArray(0);
    _vertexBuffer.release();
    _normalBuffer.release();
    _colorBuffer.release();
    _indexBuffer.release();
}

/*! Determines if the mesh is loaded and can be displayed. */
bool SurfaceMeshSphere::isLoaded(){
    return loaded;
}

/*! Changes the color to the given array. */
void SurfaceMeshSphere::changeColor(QVector<QVector3D> colors) {
    _colors = colors;
    _colorBuffer.create();
    _colorBuffer.bind();
    _colorBuffer.allocate(_colors.data(), static_cast<int>(_colors.size()) * sizeof(QVector3D));
    _colorBuffer.release();

}

/*! Returns the unit vertices. */
QVector<QVector3D> *SurfaceMeshSphere::getVerticesSphere()
{
    return &_verticesSphere;
}

/*! Returns the normals. */
QVector<QVector3D> *SurfaceMeshSphere::getNormalsSphere()
{
    return &_normalsSphere;
}

/*! Returns the colors. */
QVector<QVector3D>* SurfaceMeshSphere::getColors() {
    return &_colors;
}

/*! Returns the UVs. */
QVector<QVector2D> *SurfaceMeshSphere::getUVs()
{
    return &_uvs;
}

/*! Returns the Longitude/Latitude. */
QVector<QVector2D> *SurfaceMeshSphere::getLongiLatis()
{
    return &_longiLati;
}

/*! Returns the indices. */
QVector<GLuint>* SurfaceMeshSphere::getIndices() {
    return &_indices;
}

/*! Returns the center. */
QVector3D SurfaceMeshSphere::getCenter() const {
    return _center;
}

/*! Returns the radius. */
float SurfaceMeshSphere::getRadius() const {
    return _radius;
}

/*! Returns the barycenter. */
QVector3D SurfaceMeshSphere::getBarycenter() const {
    return _barycenter;
}
