#include "meshview.h"

/*! Constructor of meshView. */
meshView::meshView(QWidget *parent) : QOpenGLWidget(parent){
  QSurfaceFormat f = QSurfaceFormat::defaultFormat();
  QSysInfo sysinfo;
  if(sysinfo.productType().toCaseFolded() == QString("osx") ||
     sysinfo.productType().toCaseFolded() == QString("macos")){
      f.setVersion( 3, 3 );
      f.setProfile(QSurfaceFormat::CoreProfile);
  }
  setFormat(f);
  QSurfaceFormat::setDefaultFormat(f);

  _modelMatrix.setToIdentity();
  _cameraPosition = QVector3D(0.0f, 0.0f, -25.0f);
  _cameraTarget = QVector3D(0.0f, 0.0f, 0.0f);
  _viewMatrix.lookAt(_cameraPosition, _cameraTarget, QVector3D(0.0f, 1.0f, 0.0f));
  _projectionMatrix.setToIdentity();
  _projectionMatrix.perspective(45.0f, this->width() / this->height(), 0.1f, 500.0f);

  this->setFocusPolicy(Qt::StrongFocus);
  this->setMouseTracking(true);
}

/*! Initialize GL. */
void meshView::initializeGL(){
  initializeOpenGLFunctions();
  _gl = this->context()->extraFunctions();

  if(!_m_program.addShaderFromSourceFile(QOpenGLShader::Vertex, ":/shaders/mesh.vert") ||
     !_m_program.addShaderFromSourceFile(QOpenGLShader::Fragment, ":/shaders/mesh.frag") ||
     !_m_program.link()) {
    qDebug() << _m_program.log();
  }

  _currentMesh = new SurfaceMesh(_gl);

  glClearColor(0.8, 0.8, 0.8, 1);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
}

/*! Paint the current content. */
void meshView::paintGL(){
    if(_currentMesh->isLoaded()){
        if(_renderMode == 1){
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        }else if(_renderMode == 2){
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        }else if(_renderMode == 3){
            glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
        }

        _m_program.bind();
        _m_program.setUniformValue("renderMode", _renderMode);
        _m_program.setUniformValue("modelMatrix", _modelMatrix);
        _m_program.setUniformValue("viewMatrix", _viewMatrix);
        _m_program.setUniformValue("projectionMatrix", _projectionMatrix);
        _currentMesh->draw();
        _m_program.release();
    }
}

/*! Resize the window. */
void meshView::resizeGL(int w, int h){
    _projectionMatrix.setToIdentity();
    _projectionMatrix.perspective(45.0f, qreal(w) / qreal(h), 0.1, 500.0);
}

/*! Handle a wheel event. */
void meshView::wheelEvent(QWheelEvent *event){
    QMatrix4x4 translationMatrix;
    translationMatrix.setToIdentity();
    translationMatrix.translate(0.0f, 0.0f, event->delta() / 25.0f);
    _viewMatrix = translationMatrix * _viewMatrix;
    update();
}

/*! Handle a mouse press event. */
void meshView::mousePressEvent(QMouseEvent *event){
    _lastPos = event->pos();
}

/*! Handle a mouse move event. */
void meshView::mouseMoveEvent(QMouseEvent *event){


    if (event->buttons() & Qt::LeftButton) {
        float dx = event->x() - _lastPos.x();
        float dy = event->y() - _lastPos.y();
        _lastPos = event->pos();

        QMatrix4x4 transformationMatrix;
        transformationMatrix.setToIdentity();
        QVector3D vector(-dy, dx, 0.0);
        vector.normalize();
        transformationMatrix.translate(_currentMesh->getCenter());
        transformationMatrix.rotate(5.0f, vector);
        transformationMatrix.translate(-_currentMesh->getCenter());
        _modelMatrix = transformationMatrix * _modelMatrix;

        emit modelMatrixChangedMesh(_modelMatrix);
    }else if(event->buttons() & Qt::RightButton){
        float dx = event->x() - _lastPos.x();
        float dy = event->y() - _lastPos.y();
        _lastPos = event->pos();
        QMatrix4x4 transformationMatrix;
        transformationMatrix.setToIdentity();
        QVector3D vector(dx/12.5, -dy/12.5, 0.0);
        transformationMatrix.translate(vector);
        _viewMatrix = transformationMatrix * _viewMatrix;
    }
    update();
}

/*! Handle a key press event. */
void meshView::keyPressEvent(QKeyEvent *ke){
    GLfloat amount = 5.0f;
    QVector3D translation;
    switch (ke->key())
    {
        case Qt::Key_W:
            translation = QVector3D(0.0f, 0.0f, amount);
            break;
        case Qt::Key_S:
            translation = QVector3D(0.0f, 0.0f, -amount);
            break;
        case Qt::Key_D:
            translation = QVector3D(-amount, 0.0f, 0.0f);
            break;
        case Qt::Key_A:
            translation = QVector3D(amount, 0.0f, 0.0f);
            break;
        case Qt::Key_Q:
            translation = QVector3D(0.0f, amount, 0.0f);
            break;
        case Qt::Key_E:
            translation = QVector3D(0.0f, -amount, 0.0f);
            break;
        case Qt::Key_1:
            _renderMode = 1;
            emit renderModeChangedMesh(_renderMode);
            break;
        case Qt::Key_2:
            _renderMode = 2;
            emit renderModeChangedMesh(_renderMode);
            break;
        case Qt::Key_3:
            _renderMode = 3;
            emit renderModeChangedMesh(_renderMode);
            break;
    }
    QMatrix4x4 translationMatrix;
    translationMatrix.setToIdentity();
    translationMatrix.translate(translation);
    _viewMatrix = translationMatrix * _viewMatrix;
    update();
}

/*! Loads the data of a mesh with the given propeties and start the initialization. */
void meshView::loadMesh(QString path, QVector<float>* temperatures, QVector<QString>* bindingSites, QVector<QString>* atomNames, QVector<QString>* resNames, QVector<int>* aminoAcid, QMap<QString, int>* hydrophobicity, QVector<QVector<int>>* connectivity){
    QString verticesFile(path + ".vert");
    QString facesFile(path + ".face");
    QString line;
    QFile vertFile(verticesFile);
    QFile faceFile(facesFile);
    if (!vertFile.open(QFile::ReadOnly | QFile::Text) || !faceFile.open(QFile::ReadOnly | QFile::Text)){
        qDebug() << "Unable to open files.";
        qDebug() << verticesFile;
        qDebug() << facesFile;
        return;
    }

    QVector<QVector3D> *vertices = new QVector<QVector3D>;
    QVector<QVector3D> *normals = new QVector<QVector3D>;
    QVector<GLuint> *indices = new QVector<GLuint>;
    QVector<GLuint> *vertexFaceNumbers = new QVector<GLuint>;
    QVector<GLuint> *vertexClosestSpheres = new QVector<GLuint>;
    QVector<GLuint> *vertexTypes = new QVector<GLuint>;
    QVector<QString> *vertexNames = new QVector<QString>;
    QVector<GLuint> *faceTypes = new QVector<GLuint>;
    QVector<GLuint> *faceNumbers = new QVector<GLuint>;

    QVector3D center, barycenter, diff;
    float radius = 0.0001f;
    float len, alpha, alphaSq;

    QTextStream vertStream(&vertFile);
    while(vertStream.readLineInto(&line)){
        QStringList pieces = line.split(" ");
        pieces.removeAll("");
        if(pieces.length() < 10) continue;

        float x = pieces.at(0).toFloat();
        float y = pieces.at(1).toFloat();
        float z = pieces.at(2).toFloat();
        QVector3D v(x, y, z);
        barycenter += v;
        vertices->push_back(v);

        // Bouncing Bubble 1
        if(normals->size() == 0)
            center = v;
        diff = v - center;
        len = diff.length();
        if (len > radius){
            alpha = len / radius;
            alphaSq = alpha * alpha;
            radius = 0.5f * (alpha + 1 / alpha) * radius;
            center = 0.5f * ((1 + 1 / alphaSq) * center + (1 - 1 / alphaSq) * v);
        }


        normals->push_back(QVector3D(pieces.at(3).toFloat(), pieces.at(4).toFloat(),pieces.at(5).toFloat()));

        vertexFaceNumbers->push_back(pieces.at(6).toInt());
        vertexClosestSpheres->push_back(pieces.at(7).toInt());
        vertexTypes->push_back(pieces.at(8).toInt());
        vertexNames->push_back(pieces.at(9));
    }

    // Bouncing Bubble 2
    for(int i = 0; i < vertices->size(); i++){
        diff = vertices->at(i) - center;
        len = diff.length();
        if (len > radius){
            alpha = len / radius;
            alphaSq = alpha * alpha;
            radius = 0.5f * (alpha + 1 / alpha) * radius;
            center = 0.5f * ((1 + 1 / alphaSq) * center + (1 - 1 / alphaSq) * vertices->at(i));
        }
    }

    // Bouncing Bubble 3
    for(int i = 0; i < vertices->size(); i++){
        diff = vertices->at(i) - center;
        len = diff.length();
        if (len > radius){
            // without normalized dir slightly better result with cube example?
            QVector3D dir = diff.normalized();
            float d = (len - radius);
            radius += d / 2.0f;
            center += dir * (d / 2.0f);
            // same results as:
            //radius = (radius + len) / 2.0f;
            //center = center + ((len - radius) / len * diff);
        }
    }

    _cameraPosition = QVector3D(center.x(), center.y(), center.z() - qMin(2.5f * radius, 250.0f));
    _cameraTarget = center;
    _viewMatrix.setToIdentity();
    _viewMatrix.lookAt(_cameraPosition, _cameraTarget, QVector3D(0.0f, 1.0f, 0.0f));
    _modelMatrix.setToIdentity();
    _renderMode = 1;

    QTextStream faceStream(&faceFile);
    while(faceStream.readLineInto(&line)){
        QStringList pieces = line.split(" ");
        pieces.removeAll("");
        if(pieces.length() < 4) continue;
        indices->push_back(pieces.at(0).toInt()-1);
        indices->push_back(pieces.at(1).toInt()-1);
        indices->push_back(pieces.at(2).toInt()-1);
        faceTypes->push_back(pieces.at(3).toInt());
        faceNumbers->push_back(pieces.at(4).toInt());
    }

    _currentMesh->set(vertices, normals, vertexFaceNumbers, vertexClosestSpheres, vertexTypes, vertexNames, indices, faceTypes, faceNumbers, temperatures, bindingSites, atomNames, resNames, aminoAcid, hydrophobicity, connectivity, center, radius, barycenter/vertices->size());
    float genus = (2.0 - (vertices->size() - (3.0/2.0) * (indices->size()/3.0) + (indices->size()/3.0))) / 2.0;
    qDebug() << "Genus: " << ceil(genus);

    if(genus > 0.0){
        QMessageBox error;
        error.setWindowFlags(Qt::SplashScreen);
        error.setText("Only Meshes with Genus 0 allowed.");
        error.exec();
    }else{
        emit newMeshLoaded(_currentMesh);
    }
}

/*! Changes the color to the selected index. */
void meshView::changeColor(int colorSelection){
    if(_currentMesh->isLoaded()) _currentMesh->changeColor(colorSelection);
    emit colorChanged(_currentMesh->getColors());
}

/*! If enabled, takes the input modelMatrix as its own. */
void meshView::modelMatrixChangedMeshSphere(QMatrix4x4 modelMatrix)
{
    if(!_combinedRotation) return;
    _modelMatrix = modelMatrix;
    update();
}

/*! Takes the input renderMode as its own. */
void meshView::renderModeChangedMeshSphere(int renderMode)
{
    _renderMode = renderMode;
    update();
}

/*! Toggles the combined rotation. */
void meshView::combinedRotationToggle(bool toggle){
    _combinedRotation = toggle;
}
