#include "surfacemesh.h"

/*! Constructor of SurfaceMesh.*/
SurfaceMesh::SurfaceMesh(QOpenGLExtraFunctions *gl){
    _gl = gl;
}

/*! Destructor of SurfaceMesh. */
SurfaceMesh::~SurfaceMesh(){
    _vertexBuffer.destroy();
    _normalBuffer.destroy();
    _colorBuffer.destroy();
    _indexBuffer.destroy();
    _gl->glDeleteVertexArrays(1, &_vao);
}

/*! Sets the current mesh with all data given. This method is used when a new molecule is loaded. */
void SurfaceMesh::set(QVector<QVector3D>* vertices, QVector<QVector3D>* normals, QVector<GLuint>* vertexFaceNumbers, QVector<GLuint>* vertexClosestSpheres, QVector<GLuint>* vertexTypes, QVector<QString>* vertexNames, QVector<GLuint>* indices, QVector<GLuint>* faceTypes, QVector<GLuint>* faceNumbers, QVector<float>* temperatures, QVector<QString>* bindingSites, QVector<QString>* atomNames, QVector<QString>* resNames, QVector<int>* aminoAcid, QMap<QString, int>* hydrophobicity, QVector<QVector<int>>* connectivity, QVector3D center, float radius, QVector3D barycenter){
    _vertices = vertices;
    _normals = normals;
    _vertexFaceNumbers = vertexFaceNumbers;
    _vertexClosestSpheres = vertexClosestSpheres;
    _vertexTypes = vertexTypes;
    _vertexNames = vertexNames;
    _indices = indices;
    _faceTypes = faceTypes;
    _faceNumbers = faceNumbers;
    _temperatures = temperatures;
    _bindingSites = bindingSites;
    _atomNames = atomNames;
    _resNames = resNames;
    _aminoAcid = aminoAcid;
    _hydrophobicity = hydrophobicity;
    _connectivity = connectivity;
    _center = center;
    _radius = radius;
    _barycenter = barycenter;
    _colors = new QVector<QVector3D>;
    getColors(_colors, _currentColor);

    //Bind VAO
    _gl->glGenVertexArrays(1, &_vao);
    _gl->glBindVertexArray(_vao);

    //Bind VBO
    //Bind vertexBuffer
    _vertexBuffer.create();
    _vertexBuffer.bind();
    _vertexBuffer.allocate(_vertices->data(), static_cast<int>(_vertices->size()) * sizeof(QVector3D));

    //Bind normalBuffer
    _normalBuffer.create();
    _normalBuffer.bind();
    _normalBuffer.allocate(_normals->data(), static_cast<int>(_normals->size()) * sizeof(QVector3D));

    //Bind colorBuffer
    _colorBuffer.create();
    _colorBuffer.bind();
    _colorBuffer.allocate(_colors->data(), static_cast<int>(_colors->size()) * sizeof(QVector3D));

    //Bind indexBuffer
    QOpenGLBuffer inedxBufferProxy(QOpenGLBuffer::IndexBuffer);
    inedxBufferProxy.create();
    _indexBuffer = inedxBufferProxy;
    _indexBuffer.bind();
    _indexBuffer.allocate(_indices->data(), static_cast<int>(_indices->size()) * sizeof(GLuint));

    loaded = true;

    _gl->glBindVertexArray(0);
    _vertexBuffer.release();
    _normalBuffer.release();
    _colorBuffer.release();
    _indexBuffer.release();
}

// These functions were just move to another file to keep this file smaller
#include "colorcreator.h"

/*! Draws the mesh. */
void SurfaceMesh::draw(){
    _gl->glBindVertexArray(_vao);
    _gl->glEnableVertexAttribArray(0);
    _vertexBuffer.bind();
    _gl->glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);

    _gl->glEnableVertexAttribArray(1);
    _normalBuffer.bind();
    _gl->glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);

    _gl->glEnableVertexAttribArray(2);
    _colorBuffer.bind();
    _gl->glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 0, 0);

    _indexBuffer.bind();
    _gl->glDrawElements(GL_TRIANGLES, _indices->size() * 3, GL_UNSIGNED_INT, 0);

    _gl->glDisableVertexAttribArray(0);
    _gl->glDisableVertexAttribArray(1);
    _gl->glDisableVertexAttribArray(2);
    _gl->glBindVertexArray(0);
    _vertexBuffer.release();
    _normalBuffer.release();
    _colorBuffer.release();
    _indexBuffer.release();
}

/*! Determines if the mesh is loaded and can be displayed. */
bool SurfaceMesh::isLoaded(){
    return loaded;
}

/*! Changes the color to the given index. */
void SurfaceMesh::changeColor(int colorSelection) {
    if(_currentColor == colorSelection) return;
    if(colorSelection == -1 && _currentColor == -1){
        _currentColor = 0;
    }else if(colorSelection != -1){
        _currentColor = colorSelection;
    }
    getColors(_colors, _currentColor);

    _colorBuffer.create();
    _colorBuffer.bind();
    _colorBuffer.allocate(_colors->data(), static_cast<int>(_colors->size()) * sizeof(QVector3D));
    _colorBuffer.release();
}

/*! Returns the vertices. */
QVector<QVector3D>* SurfaceMesh::getVertices() const {
    return _vertices;
}

/*! Returns the normals. */
QVector<QVector3D>* SurfaceMesh::getNormals() const {
    return _normals;
}

/*! Returns the colors. */
QVector<QVector3D>* SurfaceMesh::getColors() const {
    return _colors;
}

/*! Returns the indices. */
QVector<GLuint>* SurfaceMesh::getIndices() const {
    return _indices;
}

/*! Returns the center. */
QVector3D SurfaceMesh::getCenter() const {
    return _center;
}

/*! Returns the radius. */
float SurfaceMesh::getRadius() const {
    return _radius;
}

/*! Returns the barycenter. */
QVector3D SurfaceMesh::getBarycenter() const {
    return _barycenter;
}
